from jinja2 import environmentfilter, Markup, Undefined
from django.utils.safestring import EscapeData, SafeData


def django_filter_to_jinja2(filter_func):
    """
    Note: Due to the way this function is used by
    ``coffin.template.Library``, it needs to be able to handle native
    Jinja2 filters and pass them through unmodified. This necessity
    stems from the fact that it is not always possible to determine
    the type of a filter.

    TODO: Django's "func.is_safe" is not yet handled
    """
    def _convert_out(v):
        if isinstance(v, SafeData):
            return Markup(v)
        if isinstance(v, EscapeData):
            return Markup.escape(v)  # not 100% equivalent, see mod docs
        return v

    def _convert_in(v):
        if isinstance(v, Undefined):
            # Essentially the TEMPLATE_STRING_IF_INVALID default
            # setting. If a non-default is set, Django wouldn't apply
            # filters. This is something that we neither can nor want to
            # simulate in Jinja.
            return ''
        return v

    def conversion_wrapper(value, *args, **kwargs):
        result = filter_func(_convert_in(value), *args, **kwargs)
        return _convert_out(result)
    # Jinja2 supports a similar machanism to Django's
    # ``needs_autoescape`` filters: environment filters. We can
    # thus support Django filters that use it in Jinja2 with just
    # a little bit of argument rewriting.
    if hasattr(filter_func, 'needs_autoescape'):
        @environmentfilter
        def autoescape_wrapper(environment, *args, **kwargs):
            kwargs['autoescape'] = environment.autoescape
            return conversion_wrapper(*args, **kwargs)
        return autoescape_wrapper
    else:
        return conversion_wrapper



from django.test import TestCase
from django.test.client import Client


class TestLoader(TestCase):

    def test_render(self):
        client = Client()
        response = client.get('/fullstack/')

        content = response.content

        self.assertTrue('my-foo-filter' in content)



"""Adapter for using Jinja2 with Django."""
import imp
import logging

import jinja2
from django.conf import settings
from django.template.base import Origin, TemplateDoesNotExist
from django.template.context import get_standard_processors
from django.template.loader import BaseLoader
from django.utils.importlib import import_module

from django_cofingo.utils import django_filter_to_jinja2


VERSION = (0, 1, 1)
__version__ = '.'.join(map(str, VERSION))

log = logging.getLogger('django_cofingo')


class Environment(jinja2.Environment):

    def __init__(self):
        self._libraries = []

        loader = jinja2.ChoiceLoader(self._get_loaders())
        options = self._get_options()

        super(Environment, self).__init__(
            extensions=options['extensions'],
            loader=loader,
            trim_blocks=True,
            autoescape=True,
            auto_reload=settings.DEBUG,
        )

        # Note: options already includes Jinja2's own builtins (with
        # the proper priority), so we want to assign to these attributes.
        self.filters = options['filters'].copy()
        self.globals = options['globals'].copy()
        self.tests = options['tests'].copy()
        for key, value in options['attrs'].items():
            setattr(self, key, value)

        # Set the environment in all libraries found.
        for library in self._libraries:
            library.set_env(self)

        # Hook Jinja's i18n extension up to Django's translation backend
        # if i18n is enabled; note that we differ here from Django, in that
        # Django always has it's i18n functionality available (that is, if
        # enabled in a template via {% load %}), but uses a null backend if
        # the USE_I18N setting is disabled. Jinja2 provides something similar
        # (install_null_translations), but instead we are currently not
        # enabling the extension at all when USE_I18N=False.
        # While this is basically an incompatibility with Django, currently
        # the i18n tags work completely differently anyway, so for now, I
        # don't think it matters.
        if settings.USE_I18N:
            from django.utils import translation
            self.install_gettext_translations(translation)

    def from_string(self, source, globals=None, template_class=None):
        return super(Environment, self).from_string(
            source, globals, template_class or Template)

    def _get_loaders(self):
        """Mimic Django's setup by loading templates from directories in
        TEMPLATE_DIRS and packages in INSTALLED_APPS.

        """
        x = ((jinja2.FileSystemLoader, settings.TEMPLATE_DIRS),
             (jinja2.PackageLoader, settings.INSTALLED_APPS))
        return [loader(p) for loader, places in x for p in places]

    def _get_options(self):
        """Generate a dictionary with the extensions, filters, globals, tests
        and attrs to be set on the Jinja2 environment.

        The options are added in the following order::

            * Django builtin filters
            * Jinja2 builtin filters, tests, and globals
            * Cofingo builtin filters, and extensions
            * Libraries from Django's INSTALLED_APPS (helpers.py)
            * Filters, extensions, tests and globals from DJANGO_SETTINGS

        """
        from django_cofingo import filters as cofingo_filters
        from django_cofingo import extensions as cofingo_extensions
        from django.conf import settings
        from django.template import builtins as django_builtins
        from django.core.urlresolvers import get_callable

        # Note that for extensions, the order in which we load the libraries
        # is not maintained: https://github.com/mitsuhiko/jinja2/issues#issue/3
        # Extensions support priorities, which should be used instead.
        options = {
            'extensions': [],
            'filters': {},
            'globals': {},
            'tests': {},
            'attrs': {}
        }

        def update_options(library):
            options['extensions'].extend(library.extensions)
            options['filters'].update(library.filters)
            options['globals'].update(library.globals)
            options['tests'].update(library.tests)
            options['attrs'].update(library.attrs)

        # Start with Django's builtins; this give's us all of Django's
        # filters courtasy of our interop layer.
        for lib in django_builtins:
            for name, func in lib.filters.iteritems():
                options['filters'][name] = django_filter_to_jinja2(func)

        # The stuff Jinja2 comes with by default should override Django.
        options['filters'].update(jinja2.defaults.DEFAULT_FILTERS)
        options['tests'].update(jinja2.defaults.DEFAULT_TESTS)
        options['globals'].update(jinja2.defaults.DEFAULT_NAMESPACE)

        # Our own set of builtins are next, overwriting Jinja2's.
        update_options(cofingo_filters.library)
        update_options(cofingo_extensions.library)

        # Optionally, include the i18n extension.
        if settings.USE_I18N:
            options['extensions'].append('jinja2.ext.i18n')

        # Next, add the globally defined extensions from the settings
        options['extensions'].extend(
            list(getattr(settings, 'JINJA2_EXTENSIONS', [])))

        # Finally, add extensions defined in application's templatetag
        # libraries
        for library in self._get_app_libraries():
            update_options(library)

        def from_setting(setting):
            retval = {}
            setting = getattr(settings, setting, {})
            if isinstance(setting, dict):
                for key, value in setting.iteritems():
                    retval[key] = callable(value) and value \
                        or get_callable(value)
            else:
                for value in setting:
                    value = callable(value) and value or get_callable(value)
                    retval[value.__name__] = value
            return retval
        options['filters'].update(from_setting('JINJA2_FILTERS'))
        options['globals'].update(from_setting('JINJA2_GLOBALS'))
        options['tests'].update(from_setting('JINJA2_TESTS'))

        for module_name in getattr(settings, 'COFINGO_HELPERS', []):
            module = import_module(module_name)
            update_options(module.library)
        return options

    def _get_app_libraries(self):
        for app in settings.INSTALLED_APPS:
            try:
                app_path = import_module(app).__path__
                imp.find_module('helpers', app_path)
            except (AttributeError, ImportError):
                continue

            module = import_module('%s.helpers' % app)
            if hasattr(module, 'library'):

                # Set the environment in the library
                self._libraries.append(module.library)
                yield module.library


def render_to_string(request, template, context=None):
    """
    Render a template into a string.
    """
    def get_context():
        c = {} if context is None else context.copy()
        for processor in get_standard_processors():
            c.update(processor(request))
        return c

    # If it's not a Template, it must be a path to be loaded.
    if not isinstance(template, jinja2.environment.Template):
        template = env.get_template(template)

    return template.render(get_context())

# Create the environment
env = Environment()


class Template(jinja2.Template):

    def render(self, context={}):
        """Render's a template, context can be a Django Context or a
        dictionary.
        """
        # flatten the Django Context into a single dictionary.
        context_dict = {}
        if hasattr(context, 'dicts'):
            for d in context.dicts:
                context_dict.update(d)
        else:
            context_dict = context

            # Django Debug Toolbar needs a RequestContext-like object in order
            # to inspect context.
            class FakeRequestContext:
                dicts = [context]
            context = FakeRequestContext()

        # Used by debug_toolbar.
        if settings.TEMPLATE_DEBUG:
            from django.test import signals
            self.origin = Origin(self.filename)
            signals.template_rendered.send(sender=self, template=self,
                                           context=context)

        # Jinja2 internally converts the context instance to a dictionary, thus
        # we need to store the current_app attribute as a key/value pair.
        context_dict['_current_app'] = getattr(context, 'current_app', None)

        return super(Template, self).render(context_dict)


class Loader(BaseLoader):
    is_usable = True
    env.template_class = Template

    def load_template(self, template_name, template_dirs=None):
        if hasattr(template_name, 'rsplit'):
            exclude_apps = getattr(settings, 'COFINGO_EXCLUDE_APPS', [
                'debug_toolbar',
                'admin'
            ])

            app = template_name.rsplit('/')[0]
            if app in exclude_apps:
                raise TemplateDoesNotExist(template_name)
        try:
            template = env.get_template(template_name)
            return template, template.filename
        except jinja2.TemplateNotFound:
            raise TemplateDoesNotExist(template_name)




from jinja2 import Environment
from django.test import TestCase


class TestLoadExtension(TestCase):

    def test_load(self):
        from django_cofingo.extensions import LoadExtension
        env = Environment(extensions=[LoadExtension])

        # the load tag is a no-op
        self.assertEqual(env.from_string('a{% load %}b').render(), 'ab')
        self.assertEqual(
            env.from_string('a{% load news.photos %}b').render(), 'ab')
        self.assertEqual(
            env.from_string('a{% load "news.photos" %}b').render(), 'ab')

        # [bug] invalid code was generated under certain circumstances
        self.assertEqual(
            env.from_string('{% set x=1 %}{% load "news.photos" %}').render(),
            '')


class TestSpacelessExtension(TestCase):

    def test_spaceless(self):
        from django_cofingo.extensions import SpacelessExtension
        env = Environment(extensions=[SpacelessExtension])

        result = env.from_string("""{% spaceless %}
                <p>
                    <a href="foo/">Foo</a>
                </p>
            {% endspaceless %}""").render()
        expected = '<p><a href="foo/">Foo</a></p>'
        self.assertEqual(result, expected)

    def test_spaceless_strong(self):
        from django_cofingo.extensions import SpacelessExtension
        env = Environment(extensions=[SpacelessExtension])

        result = env.from_string("""{% spaceless %}
                <strong>
                    Hello
                </strong>
            {% endspaceless %}""").render()
        expected = u'<strong>\n                    Hello' \
                    '\n                </strong>'
        self.assertEqual(result, expected)


class TestUrlExtension(TestCase):
    def setUp(self):
        from django_cofingo.extensions import URLExtension

        self.env = Environment(extensions=[URLExtension])

    def _from_string(self, source):
        from django_cofingo import Template
        return self.env.from_string(source, template_class=Template)

    def test_url(self):
        from jinja2.exceptions import TemplateSyntaxError
        from django.core.urlresolvers import NoReverseMatch

        for template, context, expected_result in (
            # various ways to specify the view
            ('{% url urls_app.views.index %}', {}, '/url_test/'),
            ('{% url apps.urls_app.views.index %}', {}, '/url_test/'),
            ('{% url "apps.urls_app.views.index" %}', {}, '/url_test/'),
            ('{% url "apps.urls_app.views.indexXX"[:-2] %}', {}, '/url_test/'),
            ('{% url the-index-view %}', {}, '/url_test/'),

            # various ways to specify the arguments
            ('{% url urls_app.views.sum 1,2 %}', {}, '/url_test/sum/1,2'),
            ('{% url urls_app.views.sum left=1,right=2 %}', {},
             '/url_test/sum/1,2'),
            ('{% url urls_app.views.sum l,2 %}', {'l': 1},
             '/url_test/sum/1,2'),
            ('{% url urls_app.views.sum left=l,right=2 %}', {'l': 1},
             '/url_test/sum/1,2'),

            # full expressive syntax
            ('{% url urls_app.views.sum left=2*3,right=z()|length %}',
                    {'z': lambda: 'u'}, '/url_test/sum/6,1'),

            # regression: string view followed by a string argument works
            ('{% url "urls_app.views.sum" "1","2" %}', {},
             '/url_test/sum/1,2'),

            # failures
            ('{% url %}', {}, TemplateSyntaxError),
            ('{% url 1,2,3 %}', {}, TemplateSyntaxError),
            ('{% url inexistant-view %}', {}, NoReverseMatch),

            # ValueError, not TemplateSyntaxError:
            # We actually support parsing a mixture of positional and keyword
            # arguments, but reverse() doesn't handle them.
            ('{% url urls_app.views.sum left=1,2 %}', {'l': 1}, ValueError),

            # as-syntax
            ('{% url urls_app.views.index as url %}', {}, ''),
            ('{% url urls_app.views.index as url %}{{url}}', {}, '/url_test/'),
            ('{% url inexistent as url %}{{ url }}', {}, ''),    # no exception
        ):
            try:
                actual_result = self._from_string(template).render(context)
            except Exception, e:
                self.assertEqual(type(e), expected_result)
            else:
                self.assertEqual(actual_result, expected_result)

    def test_url_current_app(self):
        """Test that the url can deal with the current_app context setting."""
        from django.template import RequestContext
        from django.http import HttpRequest

        template = self._from_string('{% url testapp:the-index-view %}')
        result = template.render(RequestContext(HttpRequest()))
        self.assertEqual(result, '/app/one/')

        result = template.render(
            RequestContext(HttpRequest(), current_app='two'))
        self.assertEqual(result, '/app/two/')


class TestWithExtension(TestCase):
    def test_with(self):
        from django_cofingo.extensions import WithExtension
        env = Environment(extensions=[WithExtension])

        template = env.from_string(
            '{{ x }}{% with y as x %}{{ x }}{% endwith %}{{ x }}')
        result = template.render({'x': 'x', 'y': 'y'})
        self.assertEqual(result, 'xyx')


class TestCacheExtension(TestCase):
    def test_cache(self):
        from django_cofingo.extensions import CacheExtension
        env = Environment(extensions=[CacheExtension])

        x = 0
        result = env.from_string(
            '{%cache 500 "ab"%}{{x}}{%endcache%}').render({'x': x})
        self.assertEqual(result, '0')

        # cache is used; Jinja2 expressions work
        x += 1
        result = env.from_string(
            '{%cache 50*10 "a"+"b"%}{{x}}{%endcache%}').render({'x': x})
        self.assertEqual(result, '0')

        # vary-arguments can be used
        x += 1
        result = env.from_string(
            '{%cache 50*10 "ab" x "foo"%}{{x}}{%endcache%}').render({'x': x})
        self.assertEqual(result, '2')

        x += 1
        result = env.from_string(
            '{%cache 50*10 "ab" x "foo"%}{{x}}{%endcache%}').render({'x': x})
        self.assertEqual(result, '3')



class Library(object):

    def __init__(self):
        self.env = None
        self.extensions = []
        self.filters = {}
        self.globals = {}
        self.tests = {}
        self.attrs = {}

    def set_env(self, env):
        self.env = env

    def filter(self, func):
        self.filters[func.__name__] = func

        if self.env:
            self.env.filters[func.__name__] = func

    def attr(self, name, value):
        self.attrs[name] = value
        if self.env:
            setattr(self.env, name, value)

    def extension(self, extension):
        self.extensions.append(extension)
        if self.env:
            self.env.add_extension(extension)



from datetime import datetime, date
from django.test import TestCase


def r(s, context={}):
    from django_cofingo import env
    return env.from_string(s).render(context)


class TestFilters(TestCase):
    def test_django_builtins_available(self):
        """Many filters have not been re-implemented specifically for
        Coffin, but instead the Django version is used through an
        interop-layer.

        Make sure that those are properly made available in Jinja2.
        """
        from django_cofingo import filters
        assert not hasattr(filters, 'get_digit')  # has no port
        assert r('{{ "23475"|get_digit("2") }}') == '7'
        assert r('{{ unknown|get_digit("2") }}') == ''

    def test_jinja2_builtins(self):
        """Ensure that the Jinja2 builtins are available, and take
        precedence over the Django builtins (which we automatically convert
        and install).
        """
        # Django's default filter only accepts one argument.
        assert r('{{ unknown|default("2", True) }}') == '2'

    def test_default(self):
        """We make the Jinja2 default filter behave like Django's without
        arguments, but still support Jinja2 extended syntax.
        """
        assert r('{{ foo|default("default") }}') == 'default'
        assert r('{{ foo|default("default") }}', {'foo': False}) == 'default'
        assert r('{{ foo|default("default", False) }}', {'foo': False}) == 'False'

    def test_pluralize(self):
        assert r('vote{{ 0|pluralize }}') == 'votes'
        assert r('vote{{ 1|pluralize }}') == 'vote'
        assert r('class{{ 2|pluralize("es") }}') == 'classes'
        assert r('cand{{ 0|pluralize("y", "ies") }}') == 'candies'
        assert r('cand{{ 1|pluralize("y", "ies") }}') == 'candy'
        assert r('cand{{ 2|pluralize("y", "ies") }}') == 'candies'
        assert r('vote{{ [1,2,3]|pluralize }}') == 'votes'
        assert r('anonyme{{ 0|pluralize("r", "") }}') == 'anonyme'
        assert r('anonyme{{ 1|pluralize("r", "") }}') == 'anonymer'
        assert r('vote{{ 1|pluralize }}') == 'vote'

        with self.assertRaises(TypeError):
            r('vote{{ x|pluralize }}', {'x': object()})
        with self.assertRaises(ValueError):
            r('vote{{ x|pluralize }}', {'x': 'foo'})

    def test_floatformat(self):
        assert r('{{ 1.3434|floatformat }}') == '1.3'
        assert r('{{ 1.3511|floatformat }}') == '1.4'
        assert r('{{ 1.3|floatformat(2) }}') == '1.30'
        assert r('{{ 1.30|floatformat(-3) }}') == '1.300'
        assert r('{{ 1.000|floatformat(3) }}') == '1.000'
        assert r('{{ 1.000|floatformat(-3) }}') == '1'
        with self.assertRaises(ValueError):
            r('{{ "foo"|floatformat(3) }}')

        with self.assertRaises(ValueError):
            r('{{ 4.33|floatformat("foo") }}')

    def test_date_stuff(self):
        from django_cofingo import env
        assert r('a{{ d|date("Y") }}b', {'d': date(2007, 01, 01)}) == 'a2007b'
        assert r('a{{ d|time("H") }}b', {'d': datetime(2007, 01, 01, 12, 01, 01)}) == 'a12b'
        # TODO: timesince, timeuntil

        # Make sure the date filters can handle unset values gracefully.
        # While generally we'd like to be explicit instead of hiding errors,
        # this is a particular case where it makes sense.
        for f in ('date', 'time', 'timesince', 'timeuntil'):
            assert r('a{{ d|%s }}b' % f) == 'ab'
            assert r('a{{ d|%s }}b' % f, {'d': None}) == 'ab'

